program NMO_mute
  ! Program for removing bubble from data
  ! Based on GEE's pef and helicon modules
  ! Calculates moving window containing reflections
  !   PEF is built based on those reflections
  !
  ! Inputs:
  !   1) seplib format data (n1 is time, n2 is number of traces)
  !   2) t0 [0.45]:  zero-offset estimate (s)
  !   3) header: history file of data that includes offsets 
  !   4) vel [1500.0]: water velocity (m/s)
  !   5) minoff [0.0] : maximum offset to consider in windowing for PEF (meters)
  !   6) offKey: The index in the provided header file that indicates the offsets
  !
  ! Outputs:
  !   1) history file of muted data
  !
  !
  !=================================
  ! Standard declarations
  !=================================
  use sep
  implicit none

  !=================================
  ! Declare variables
  !=================================
  ! Program variables
  integer                           :: ii,it,nt,ntraces,offKey,nKeys
  real                              :: tt,vel,t0,minoff,o1,d1,dz,off,d2,o2
  real, dimension(:),allocatable    :: offset,input,output,sz,gz
  real, dimension(:,:),allocatable  :: header

  !=================================
  ! Open I/O
  !=================================
  call sep_init()

	! Read history file parameters
	call from_history('n1',nt,1)
	call from_history('n2',ntraces,1)
	call from_history('o1',o1,0.)
	call from_history('d1',d1,1.)

	! Write history file parameters
	call to_history('n1',nt)
	call to_history('n2',ntraces)
	call to_history('o1',o1)
	call to_history('d1',d1)

  !=================================
  ! Initialization
  !=================================  

  ! Window parameters
  call from_param('dz',dz,25.0)
  call from_param('vel',vel,1500.)
  call from_param('minoff',minoff,0.0)

  ! Allocate arrays
  allocate(input(nt))
  allocate(output(nt))

  ! Read in header
  if (exist_file('header')) then
		call from_aux("header","n1",nKeys)
		call from_param('offKey',offKey,8)
		allocate(sz(ntraces),gz(ntraces),offset(ntraces))
		allocate(header(nKeys,ntraces))
		call sep_read(header,'header')
		! Grab trace offset info
		offset=header(offKey,:)
		sz=header(2,:)
		gz=header(5,:)
  else
  		write(0,*)"Assuming that input is a regularized file."
		call from_param('t0',t0,0.5)
		call from_history('o2',o2,0.)
		call from_history('d2',d2,25.0)
  endif

  !=================================
  ! Kernel
  !=================================

  ! Window according to parabola
	if (exist_file('header')) then
		do ii=1,ntraces
			write(0,*) "Percent done grabbing window: ", 100*ii/ntraces
			t0=(sz(ii)-gz(ii))*dz/vel
			if (offset(ii)>=minoff) then ! Ignore the small offsets if minoff set
				tt = sqrt( (t0*t0) + ((offset(ii)*offset(ii))/(vel*vel)) ) ! Window according to a hyperbola
				it = 1.5+(tt-o1)/d1
				call sep_read(input)
				output = input
				output(it:nt) = 0.0 ! Zero everything below the direct arrival
			else
				output=0.0
			endif
			call sep_write(output)
		end do
	else ! Assuming a regularized file
		do ii=1,ntraces
			write(0,*) "Percent done grabbing window: ", 100*ii/ntraces
			off = o2+d2*(ii-1)
			write(0,*) "off = ", off
			if (off>=minoff) then ! Ignore the small offsets if minoff set
				tt = sqrt( (t0*t0) + (off*off)/(vel*vel)) ! Window according to a hyperbola
				it = 1.5+(tt-o1)/d1
				write(0,*) "it = ", it
				call sep_read(input)
				output = input
				output(it:nt) = 0.0 ! Zero everything below the direct arrival
			else
				output=0.0
			endif
			call sep_write(output)
		end do
	end if




  ! Deallocate arrays
  ! deallocate(input,output,header,offset)
  write(0,*) " DONE!"

end program




